////////////////////////////////////////////////////////////////////////////////
//
#define TITLE "bin2iso - Convert CD .BIN to .ISO"
#define COPYR "Copyright (C) 2011 Neill Corlett"
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
////////////////////////////////////////////////////////////////////////////////

#include "common.h"
#include "banner.h"
#include "tinydir.h"

////////////////////////////////////////////////////////////////////////////////
//
// Sector buffer
//
uint8_t buffer[2352];

////////////////////////////////////////////////////////////////////////////////

int fileconvert(char* filename) {
    int returncode = 0;
    char* binfilename = NULL;
    char* isofilename = NULL;
    FILE* binfile = NULL;
    FILE* isofile = NULL;
    time_t tstart = time(NULL) - 1;
    off_t binfilelength = 0;
    off_t binfilepos = 0;
    char warning_shown_sync  = 0;
    char warning_shown_mode  = 0;
    char warning_shown_form2 = 0;

    //
    // Process command line
    //
    binfilename = filename;
    isofilename = str_replace(binfilename, ".bin", ".iso");
    if(!strcmp(binfilename, isofilename)) {
        isofilename = (char *)calloc(strlen(binfilename)+5, sizeof(char));
        sprintf(isofilename, "%s.iso", binfilename);
    }

    //
    // Make sure the iso doesn't already exist
    //
    {   int exists = 0;
        FILE* test = fopen(isofilename, "rb");
        exists = (test != NULL);
        if(test) {
            fclose(test);
        }
        if(exists) {
            printf("Warning: %s already exists.\n", isofilename);
            printf("Overwriting.\n");
        }
    }

    //
    // Open bin for reading
    //
    binfile = fopen(binfilename, "rb");
    if(!binfile) { goto error_bin; }
    //
    // Verify the file length
    //
    if(fseeko(binfile, 0, SEEK_END) != 0) { goto error_bin; }
    binfilelength = ftello(binfile);
    if(!binfilelength) {
        printf("Error: %s: Empty file; nothing to convert.\n", binfilename);
        goto error_file;
    }
    if((binfilelength % ((off_t)2352)) != 0) {
        printf("Warning: %s: File size is not a multiple of 2352.\n", binfilename);
        printf("Likely not a BIN file.\n");
    }

    //
    // Open iso for writing
    //
    isofile = fopen(isofilename, "wb");
    if(!isofile) { goto error_iso; }

    //
    // Convert
    //
    if(fseeko(binfile, 0, SEEK_SET) != 0) { goto error_bin; }
    for(binfilepos = 0;; binfilepos += 2352) {
        size_t data_ofs = 0x10;
        size_t s;
        time_t t = time(NULL);

        //
        // Progress indicator
        //
        if(t != tstart) {
            off_t a = binfilepos    / ((off_t)2352);
            off_t b = binfilelength / ((off_t)2352);
            off_t percent = b ? ((((off_t)100) * a) / b) : 0;
            fprintf(stderr, "%d%%...\r", (int)percent);
            tstart = t;
        }

        s = fread(buffer, 1, 2352, binfile);
        if(ferror(binfile)) { goto error_bin; }
        if(s == 0 && feof(binfile)) {
            // Done converting
            break;
        }
        if(s > 2352) { s = 2352; }
        if(s < 2352) {
            printf("Warning: %s: Last sector is incomplete\n", binfilename);
            // Zero-fill the remaining data
            memset(buffer + s, 0, 2352 - s);
        }
        //
        // Check for sync
        //
        if(memcmp(buffer, "\x00\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\x00", 12)) {
            if(!warning_shown_sync) {
                printf("Warning: %s: Sync area missing at 0x", binfilename);
                fprinthex(stdout, binfilepos, 1);
                printf("\n");
                printf("Likely not a BIN file.\n");
            }
            warning_shown_sync = 1;
        }
        //
        // Determine sector mode
        //
        switch(buffer[0xF]) {
        case 0x00:
            //
            // Mode 0: technically legal, all zeroes
            //
            data_ofs = 0x10;
            break;

        case 0x01:
            //
            // Mode 1: ok
            //
            data_ofs = 0x10;
            break;

        case 0x02:
            //
            // Mode 2: Check form
            //
            data_ofs = 0x18;
            if(
                ((buffer[0x12] & 0x20) != 0) ||
                ((buffer[0x16] & 0x20) != 0)
            ) {
                if(!warning_shown_form2) {
                    printf("Warning: %s: Form 2 sector at 0x", binfilename);
                    fprinthex(stdout, binfilepos, 1);
                    printf("\n");
                    printf("Information will be lost.\n");
                }
                warning_shown_form2 = 1;
            }
            break;

        default:
            //
            // Unknown mode
            //
            if(!warning_shown_mode) {
                printf("Warning: %s: Unknown sector mode 0x%x at 0x", binfilename, buffer[0xF]);
                fprinthex(stdout, binfilepos, 1);
                printf("\n");
            }
            data_ofs = 0x10;
            if(!warning_shown_mode) {
                printf("Assuming mode 1.\n");
            }
            warning_shown_mode = 1;
            break;
        }
        //
        // Write sector to ISO
        //
        s = fwrite(buffer + data_ofs, 1, 2048, isofile);
        if(ferror(isofile)) { goto error_iso; }
        if(s != 2048) { goto error_iso; }
    }
    printf("%s -> %s done\n", binfilename, isofilename);

    if(binfile) { fclose(binfile); }
        if(isofile) {
        fclose(isofile);
        // If there was an error, remove the partial ISO file
        if(returncode && isofilename) {
            remove(isofilename);
        }
    }

    goto done_file;

error_bin: printfileerror(binfile, binfilename); goto error_file;
error_iso: printfileerror(isofile, isofilename); goto error_file;

error_file:
    returncode = 1;

done_file:
    return returncode;
}

////////////////////////////////////////////////////////////////////////////////

int dirloop(char* dirname) {
    tinydir_dir dir;
    tinydir_open_sorted(&dir, dirname);

    for (unsigned int i = 0; i < dir.n_files; i++) {
        tinydir_file file;
        tinydir_readfile_n(&dir, &file, i);

        if(!strcmp(file.name, ".") || !strcmp(file.name, "..")) { continue; }

        if(file.is_dir) {
            dirloop(file.path);
        } else {
            fileconvert(file.path);
        }
        if(i != dir.n_files-1) { printf("\n"); };
    }

    tinydir_close(&dir);

    return 0;
}

////////////////////////////////////////////////////////////////////////////////

int main(int argc, char** argv) {
    int returncode = 0;

    banner();

    normalize_argv0(argv[0]);

    if(argc == 1) { goto usage; }

    for(int i = 1; i < argc; i++) {
        if(is_directory(argv[i])) {
            dirloop(argv[i]);
        } else {
            fileconvert(argv[i]);
        }
        if(i != argc-1) { printf("\n"); };
    }

    goto done;

usage:
    printf(
        "Usage: %s input_files\n",
        argv[0]
    );
    goto error;

error:
    returncode = 1;

done:
    return returncode;
}

////////////////////////////////////////////////////////////////////////////////
